# auto -*- coding: utf-8 -*-

import argparse
import getpass
import logging
import os
import re
import sqlite3
import string
import sys
import textwrap
import subprocess
from ConfigParser import ConfigParser
from math import ceil

from colorama import Fore, Style
from requests import ConnectionError
from requests.packages import urllib3
from sbgrid import __version__ as VERSION
from sbgrid import APP_NAME, CLI_NAME, ORG_NAME
from sbgrid.core import (DamagedInstallationException, Installation,
                         NotActivatedException, RSyncPackageNotFoundError, DependenciesNotAvailableError)

urllib3.disable_warnings()


def get_tty_size():
    """Get terminal size.

    Return:
        (int, int): height, width of terminal.
    """
    return map(int, os.popen('stty size', 'r').read().split())


def tabulate(packages, width):
    max_length = max(map(lambda p: len(p.slug), packages))
    number_of_columns = int(float(width + 1) / (max_length + 1))
    if number_of_columns == 0:
        # TODO !!! Test on bch?
        return list(packages[0].slug)
    chunk = int(ceil(float(len(packages)) / number_of_columns))
    columns = [packages[i:i + chunk] for i in range(0, len(packages), chunk)]
    lines = list()
    for i in range(0, chunk):
        line = list()
        for column in columns:
            if len(column) > i:
                if column[i].installed:
                    line.append(green(column[i].slug.ljust(max_length)))
                else:
                    line.append(column[i].slug.ljust(max_length))
        lines.append(' '.join(line))
    return lines


def green(s):
    return Fore.GREEN + Style.BRIGHT + s + Style.RESET_ALL


def red(s):
    return Fore.RED + Style.BRIGHT + s + Style.RESET_ALL


def yellow(s):
    return Fore.YELLOW + Style.BRIGHT + s + Style.RESET_ALL


def blue(s):
    return Fore.BLUE + Style.BRIGHT + s + Style.RESET_ALL


def magenta(s):
    return Fore.MAGENTA + Style.BRIGHT + s + Style.RESET_ALL


class Installer(object):

    def __init__(self):
        if os.geteuid() == 0:
            sys.exit("Warning! Application can not be run as a root.")
        parser = argparse.ArgumentParser(prog=CLI_NAME, description=APP_NAME, usage='''
{0} <command> [<command_args>]

Available commands are:

Query packages:
    collections     list available collections
    grids           list available grids
    info            get detailed information on a particular package
    installed       list all installed packages
    list            list available packages
    search          list and search installed packages
    updates         list available updates

Manage packages:
    install             install packages
    install-all
    rebuild             rebuild installation of all installed packages
    remove / uninstall  remove installed packages
    resume              resume interrupted installations
    update              update installed package
    update-all

Environment:
    shell    Load {0} environment

Installation management
    activate      activate new installation
    admin         cron mode
    autoclean     removes unused packages
    reactivate    reactivate installation using existing config and package database/list
    save          save current configuration

Additional arguments:
    -h, --help              show usage information and exit current platform
    -l, --info-versions     list available and default versions for <program> on current platform
    -L, --info-all-versions list available and default versions for <program> on all platforms
    --debug                 enable verbose logging
'''.format(CLI_NAME), add_help=False)
        parser.add_argument('-h', '--help', help='show usage infromation and exit', action='store_true', default=False)
        parser.add_argument('-l', '--info-versions', help='list available and default versions for <program> on current platform', action='store_true', default=False)
        parser.add_argument('-L', '--info-all-versions', help='list available and default versions for <program> on all platforms', action='store_true', default=False)
        parser.add_argument('--debug', help='enable verbose logging', action='store_true', default=False)
        parser.add_argument('command', help='subcommand to run')
        (args, command_args) = parser.parse_known_args()
        if args.debug:
            logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
        else:
            logging.basicConfig(stream=sys.stdout, level=logging.INFO, format='%(message)s')
        for handler in logging.root.handlers:
            handler.addFilter(logging.Filter('root'))
        args.command = args.command.replace('-', '_')

        if args.command == 'version':
            logging.info('{0} command line installer version {1}'.format(CLI_NAME, VERSION))
            sys.exit(0)
        # diffrent behavior when use -l or -L flag
        if args.info_versions or args.info_all_versions:
            argument_slug = args.command
            args.command = 'info'

        if not hasattr(self, args.command):
            logging.info('Unrecognized command')
            parser.print_help()
            sys.exit(1)

        if args.help:
            command_args.append('--help')

        self.height, self.width = 25, 80
        if sys.stdin.isatty():
            self.height, self.width = get_tty_size()
        self.installation = Installation()
        self.installation.debug = args.debug

        try:
            logging.debug('Loading configuration...')
            self.installation.load()
            if args.info_all_versions:
                self.info(['--all-versions', argument_slug])
                sys.exit(1)
            if args.info_versions:
                self.info(['--versions', argument_slug])
                sys.exit(1)
        except ConnectionError:
            logging.fatal('Need Internet connection to continue! Exiting...')
            sys.exit(1)
        except NotActivatedException:
            if args.command not in ('activate', 'reactivate'):
                logging.error('Installation was not activated!')
                sys.exit(1)
        except DamagedInstallationException, e:
            logging.error('''{0} Installation is damaged! {1}'''.format(ORG_NAME, str(e)))
            sys.exit(1)

        try:
            getattr(self, args.command)(command_args)
        except ConnectionError:
            logging.fatal('Need Internet connection to continue! Exiting...')
            sys.exit(1)
        except Exception, e:
            logging.error(e.message, exc_info=args.debug)

    def _validate_grid_name(self, grid):
        for g in self.installation.grids:
            if grid.lower() == g.lower():
                return g
        raise Exception('Unknown grid name!')

    def _validate_collection_names(self, grid, collection):
        if collection and collection != 'all':
            name = ' '.join(collection).strip().lower()
            for c in self.installation.collections[grid]:
                if c.lower() == name.lower():
                    return ((grid, [c]),)
            raise Exception('Unknown collection name!')
        return ((grid, self.installation.collections[grid]), )

    def bootstrap_installer(self):
        app_name = self.installation.bootstrap_installer()
        if app_name and sys.platform == 'darwin':  # if macOs create link in Applications folder
            # TODO write test to check if link for installer was created
            app_uri = os.path.join([self.installation.link, self.installation.platform, '{0}_installer'.format(CLI_NAME), 'latest', app_name])
            app_shortcut_path = '/Applications/{0}'.format(app_name)
            if not os.path.isfile(app_shortcut_path):
                subprocess.check_output(['/bin/ln', '-s', '/'.join(app_uri), app_shortcut_path])
            else:
                logging.info('')
                logging.info('The latest version of the {0} has been installed. '
                             'Update of shortcut in Applications folder is required. [Y/n])'.format(app_name))
                if not raw_input().strip().lower().startswith('y'):
                    subprocess.check_output(['/bin/rm', app_shortcut_path])
                    subprocess.check_output(['/bin/ln', '-s', '/'.join(app_uri), app_shortcut_path])

    def activate(self, command_args):
        parser = argparse.ArgumentParser(prog='{0} activate'.format(CLI_NAME), description='Activate new installation.')
        parser.add_argument('site', help='site name')
        parser.add_argument('user', help='user name')
        parser.add_argument('key', help='activation key')
        parser.add_argument('--multibranch', help='enable installation for all architectures', action='store_true', default=False)
        args = parser.parse_args(command_args)
        if self.installation.activate(args.site, args.user, args.key):
            logging.info('Activation successfull!')
            self.installation.save_activation_info()
            if args.multibranch:
                platforms = self.installation.save_multiarch_config()
                logging.info('Multibranch installation mode is on')
                logging.info('List of available architectures: ')
                for platform_name in platforms:
                    logging.info(' - {0}'.format(platform_name))
            logging.info('Updating to latest version ...')
            self.bootstrap_installer()
        else:
            logging.info('Activation failed!')

    def list(self, command_args):
        parser = argparse.ArgumentParser(prog='{0} list'.format(CLI_NAME), description='List packages.')
        parser.add_argument('--grid', help='grid name', default=CLI_NAME)
        parser.add_argument('collection', nargs='*', help='collection name')
        args = parser.parse_args(command_args)
        logging.debug(args)
        for grid, collections in self._validate_collection_names(self._validate_grid_name(args.grid), args.collection):
            for collection in collections:
                logging.info(' -- {0} {1}'.format(magenta(collection), (self.width - len(collection) - 5) * '-'))
                for line in tabulate(self.installation.packages_by_collection(grid, collection), self.width):
                    logging.info(line)

    def admin(self, command_args):
        parser = argparse.ArgumentParser(prog='{0} admin'.format(CLI_NAME), description='cron mode')
        parser.add_argument('--all-versions', help='install all versions for a package', action='store_true')
        args = parser.parse_args(command_args)
        self.install_all(['--yes', '--all-versions' if args.all_versions else None])
        self.update_all(['--yes'])
        self.autoclean(['--yes'])

    def search(self, command_args):
        parser = argparse.ArgumentParser(prog='{0} search'.format(CLI_NAME), description='List packages that contains given query string')
        parser.add_argument('query', help='query string')
        args = parser.parse_args(command_args)
        packages = self.installation.search_packages(args.query)
        if len(packages) == 0:
            logging.info('I could not find any package for the query')
            sys.exit(1)
        for line in tabulate(packages, self.width):
            logging.info(line)

    def install_package(self, slug, all_versions):
        try:
            cmd_args = slug.split('@')
            if all_versions:
                package = self.installation.package(cmd_args[0])
                if self.installation.is_multiarch_mode:
                    for version in package.all_versions:
                        self.installation.install_with_dependencies(package, version, version.os_dir)
                else:
                    for version in package.versions:
                        self.installation.install_with_dependencies(package, version, self.installation.platform)
                return True
            else:
                if len(cmd_args) == 1:
                    package = self.installation.package(cmd_args[0])
                    if self.installation.is_multiarch_mode:
                        for platform_dir in self.installation.get_multiarch_configuration():
                            version = package.get_default_version_for_platform(platform_dir)
                            if version is None:
                                logging.info('There is no default version for {0} for {1}'.format(package.slug, platform_dir))
                            else:
                                self.installation.install_with_dependencies(package, version, version.os_dir)
                    else:
                        version = package.default_version
                        self.installation.install_with_dependencies(package, version, self.installation.platform)
                    return True
                elif len(cmd_args) == 2:
                    package = self.installation.package(cmd_args[0])
                    if self.installation.is_multiarch_mode:
                        for platform_dir in self.installation.get_multiarch_configuration():
                            version = package.get_version_for_platform(cmd_args[1], platform_dir)
                            if version is None:
                                logging.info('There is no version {0} of {1} for {2}'.format(cmd_args[1], package.slug, platform_dir))
                            else:
                                self.installation.install_with_dependencies(package, version, platform_dir)
                    else:
                        version = package.version(cmd_args[1])
                        self.installation.install_with_dependencies(package, version, self.installation.platform)
                    return True
                else:
                    raise Exception('Incorrect argument')
        except RSyncPackageNotFoundError:
            logging.info('Error: {0} is available but was not found on the download server.'
                         ' Please try again in a few minutes and if the issue persists,'
                         ' please let us know - bugs@sbgrid.org.'.format(red('{0}@{1} {2}'.format(package.slug, version.version, version.os_dir))))
        except DependenciesNotAvailableError, err:
            for missing_package, missing_version in err.not_available_dependencies:
                logging.info('Error: while installing {0} - {1} is required as a dependency,'
                             ' but was not found in your installation.'
                             ' Please see https://sbgrid.org/software/ for licensing information.'
                             .format(yellow('{0}@{1} {2}'.format(package.slug, version.version, version.os_dir)),
                                     red('{0}@{1}'.format(missing_package, missing_version))))
        except Exception, err:
            logging.error(err.message)
        return False

    def install(self, command_args):
        parser = argparse.ArgumentParser(prog='{0} install'.format(CLI_NAME), description='Install package.')
        parser.add_argument('--all-versions', help='install all versions for a package', action='store_true')
        parser.add_argument('title', help='package name or package_name@version', nargs='+')
        args = parser.parse_args(command_args)
        installation_changed = False
        for package_name in args.title:
            installation_changed = self.install_package(package_name, args.all_versions) or installation_changed
        if installation_changed:
            self.installation.setup()

    def install_all(self, command_args):
        parser = argparse.ArgumentParser(prog='{0} install-all'.format(CLI_NAME),
                                         description='Install all packages in selected collection.')
        parser.add_argument('--yes', help='continue without confirmation', action='store_true')
        parser.add_argument('--grid', help='grid name', default=CLI_NAME)
        parser.add_argument('--all-versions', help='install all versions for a package', action='store_true')
        parser.add_argument('collection', nargs='*', help='collection name', default='all')
        args = parser.parse_args(command_args)
        logging.debug(args)
        packages_to_install = list()
        for grid, collections in self._validate_collection_names(self._validate_grid_name(args.grid), args.collection):
            for collection in collections:
                for package in self.installation.packages_by_collection(grid, collection):
                        packages_to_install.append(package)
        if not args.yes:
            logging.info('Packages to be installed:')
            for package in packages_to_install:
                logging.info('%s@%s', package, package.default_version)
            logging.info('Continue? [Y/n]?')
            if not raw_input().strip().lower().startswith('y'):
                return
        for package in packages_to_install:
            self.install_package(package.slug, args.all_versions)
        if packages_to_install:
            self.installation.setup()

    def save(self, command_args):
        parser = argparse.ArgumentParser(prog='{0} save'.format(CLI_NAME), description='Save current configuration.')
        parser.add_argument('filename', help='config file', default='installer.config')
        args = parser.parse_args(command_args)
        try:
            self.installation.save_configuration(args.filename, None)
            logging.info('Current configuration stored in file {0} successfully.'.format(green(args.filename)))
        except Exception:
            logging.error('An error occurred when writing current configuration to a file!')

    def remove_package(self, slug):
        package_to_remove = slug.split('@')
        if len(package_to_remove) == 1:
            package = self.installation.package(package_to_remove[0])
            if self.installation.is_multiarch_mode:
                for platform_dir in self.installation.get_multiarch_configuration():
                    self.installation.remove_package(package, platform_dir)
                    return True
            else:
                self.installation.remove_package(package, self.installation.platform)
                return True
        elif len(package_to_remove) == 2:
            package = self.installation.package(package_to_remove[0])
            version_string = package_to_remove[1]
            if self.installation.is_multiarch_mode:
                for platform_dir in self.installation.get_multiarch_configuration():
                    version = package.get_version_for_platform(version_string, platform_dir)
                    if version:
                        self.installation.remove(package, version, version.os_dir)
                        return True
                    else:
                        logging.info('{0} for {1} is not installed!'.format(slug, platform_dir))
            else:
                version = package.version(version_string)
                self.installation.remove(package, version, version.os_dir)
                return True
        else:
            raise Exception('Incorrect argument')
        return False

    def remove(self, command_args):
        parser = argparse.ArgumentParser(prog='{0} remove'.format(CLI_NAME), description='Remove package.')
        parser.add_argument('title', help='package name or package_name@version', nargs='+')
        args = parser.parse_args(command_args)
        installation_setup = False
        for package_name in args.title:
            if self.remove_package(package_name):
                logging.info('Package {0} removed.'.format(package_name))
                installation_setup = True
        if installation_setup:
            self.installation.setup()

    def resume(self, command_args):
        parser = argparse.ArgumentParser(prog='{0} resume'.format(CLI_NAME),
                                         description='Resume interrupted installations.')
        parser.parse_args(command_args)
        logging.info('Resuming interrupted installations...')
        for slug, version in self.installation.pending_packages():
            package = self.installation.package(slug)
            logging.info('Installing: {0} ({1})'.format(slug, version))
            self.installation.install_with_dependencies(package, package.version(version))
        self.installation.setup()

    def installed(self, command_args):
        def show_legend():
            logging.info('')
            logging.info('status : {0}  {1}  {2}'
                         .format(green('installed'), blue('update available'), red('pending removal')))
        parser = argparse.ArgumentParser(prog='{0} installed'.format(CLI_NAME), description='List installed packages.')
        parser.parse_args(command_args)
        if not self.installation.installed_packages():
            return
        logging.info('Installed packages:')
        logging.info('')
        for package in self.installation.installed_packages():
            for installed_version in filter(lambda v: v.installed, package.versions):
                if installed_version.update:
                    logging.info('{0}{1}'.format(package.slug, blue('@' + installed_version.version)))
                elif installed_version.removed:
                    logging.info('{0}{1}'.format(package.slug, red('@' + installed_version.version)))
                else:
                    logging.info('{0}{1}'.format(package.slug, green('@' + installed_version.version)))

        show_legend()

    def info(self, command_args):
        def format_long_text(s):
            s = s.split('</p>')
            s = map(string.strip, s)
            s = map(lambda a: a.replace('</p>', '\n'), s)
            s = map(lambda a: a.replace('\r', ''), s)
            s = map(lambda a: re.sub(r'<[^>]*>', '', a), s)
            s = map(lambda a: textwrap.fill(a, width=75, initial_indent=5 * ' '), s)
            return '\n'.join(s)

        parser = argparse.ArgumentParser(prog='{0} info'.format(CLI_NAME), description='Print package metadata.')
        parser.add_argument('title', help='package name')
        parser.add_argument('--versions', help='list available and default versions for <program> on current platform', action='store_true')
        parser.add_argument('--all-versions', help='list available and default versions for <program> on all platforms', action='store_true')
        args = parser.parse_args(command_args)
        try:
            package = self.installation.package(args.title)
            if args.all_versions:
                versions_dict = package.get_all_versions_dict()
                for platform in versions_dict:
                    logging.info(platform + ':')
                    for version in versions_dict[platform]:
                        version_info = ' -{0}'.format(version.version)
                        if version.default:
                            version_info = version_info + ' (default)'
                        logging.info(version_info)
            elif args.versions:
                for version in package.versions:
                    version_info = ''
                    if version.installed:
                        version_info = green(version.version)
                    else:
                        version_info = version.version
                    if version.default:
                        version_info = version_info + ' (default)'
                    logging.info(version_info)
            else:
                logging.info('Package: %s', package.title)
                logging.info('Name slug: %s', package.slug)
                logging.info('\nAvailable versions:')
                for version in package.versions:
                    extra_info = []
                    if version.installed:
                        extra_info.append('installed')
                    if version.default:
                        extra_info.append('default')
                    if version.obsolete:
                        extra_info.append('obsolete')
                    extra_info = extra_info and '({0})'.format(', '.join(extra_info)) or ''
                    dependencies = []
                    for p, v in version.depends_on:
                        dependencies.append('{0} ({1})'.format(p, v))
                    dependencies = dependencies and 'depends on: {0}'.format(', '.join(dependencies)) or ''
                    logging.info('\t* %s', ' '.join([a for a in (version.version, extra_info, dependencies) if a]))
                logging.info('\nGrids(s): %s', ', '.join(package.grids))
                if package.collections:
                    logging.info('Collection(s): %s', ', '.join(package.collections))
                if package.description:
                    logging.info('\nDescription:')
                    logging.info(format_long_text(package.description))
                if package.technical_notes:
                    logging.info('\nTechnical notes:')
                    logging.info(format_long_text(package.technical_notes))
                if package.links:
                    logging.info('\nLinks:')
                    for name, link in package.links:
                        logging.info('\t{0}: {1}'.format(name.replace('_', ' ').capitalize(), link))
        except Exception, e:
            logging.error(e.message)

    def autoclean(self, command_args):
        parser = argparse.ArgumentParser(prog='{0} autoclean'.format(CLI_NAME), description='Removes unused packages.')
        parser.add_argument('--yes', help='continue without confirmation', action='store_true')
        args = parser.parse_args(command_args)
        packages_to_remove = self.installation.get_packages_to_remove()
        if not packages_to_remove:
            if not args.yes:
                logging.info('There are no packages to remove.')
            sys.exit(0)
        if not args.yes:
            logging.info('Found {0} packages to remove'.format(len(packages_to_remove)))
            for key, package in packages_to_remove.items():
                for version in filter(lambda v: v.autoclean, package.versions):
                    logging.info('{0}@{1} for {2}'.format(package.slug, version, version.os_dir))
            logging.info('')
            logging.info('Are you sure you want to remove all those packages [Y/n]')
            if not raw_input().strip().lower().startswith('y'):
                sys.exit(1)
        for key, package in packages_to_remove.items():
            for version in filter(lambda v: v.autoclean, package.versions):
                self.installation.remove(package, version, version.os_dir)
                logging.info('Package {0}@{1} removed!'.format(package.slug, version))
        self.installation.setup()

    def shell(self, command_args):
        def user_shell():
            if os.environ.get('SHELL', None):
                return os.environ.get('SHELL')
            if sys.platform == 'darwin':
                return subprocess.check_output(['/usr/bin/dscl', '.', '-read', '/Users/{0}'.format(getpass.getuser()),
                                                'UserShell', '|', 'awk', '{print $2}']).strip()
            if sys.platform.startswith('linux'):
                return subprocess.check_output(['/usr/bin/getent', 'passwd', '$LOGNAME']).split(':')[-1]
            raise Exception('Unsupported platform {0}!'.format(sys.platform))

        parser = argparse.ArgumentParser(prog='{0} shell'.format(
            CLI_NAME), description='Start {0} shell'.format(CLI_NAME))
        parser.parse_args(command_args)
        if not self.installation.installed_packages():
            logging.info('No packages installed. Exiting!')
            sys.exit(1)
        user_shell = user_shell()
        suffix = 'shrc'
        if user_shell.endswith('csh'):
            suffix = 'cshrc'
        shrc = self.installation._share_path('grids', '.admin', 'allgrid.{0}'.format(suffix))
        os.execv(user_shell, [user_shell, '--rcfile', shrc])

    def updates(self, command_args=None):
        parser = argparse.ArgumentParser(prog='{0} updates'.format(CLI_NAME),
                                         description='List available updates for installed packages.')
        parser.parse_args(command_args)
        for package in self.installation.updated_packages():
            for version in package.updated_versions:
                if version.new_version:
                    logging.info('{0} {1} -> {2}'.format(package.slug, version.version, version.new_version.version))
                else:
                    logging.info('{0} {1} {2} -> {3}'.format(package.slug, version.version,
                                                             version.installed_datetime, version.timestamp_datetime))

    def update(self, command_args):
        parser = argparse.ArgumentParser(prog='{0} upgrade'.format(CLI_NAME),
                                         description='Update package to latest release/version.')
        parser.add_argument('--yes', help='continue without confirmation', action='store_true')
        parser.add_argument('title', help='package name')
        args = parser.parse_args(command_args)
        package = self.installation.package(args.title)
        installation_changed = False
        if not package.installed:
            logging.fatal('Package %s is not installed!', args.title)
            sys.exit(1)
        if not package.update:
            logging.fatal('No updates for %s!', args.title)
            sys.exit(1)
        for version in package.updated_versions:
            if not args.yes:
                if version.new_version:
                    logging.info('Package %s %s will be upgraded to version %s. ', package.slug,
                                 version.version, version.new_version.version)
                else:
                    logging.info('Package %s %s %s will be upgraded to new release %s. ', package.slug,
                                 version.version, version.installed_datetime, version.timestamp_datetime)
                logging.info('Continue? [Y/n]?')
                if not raw_input().strip().lower().startswith('y'):
                    continue
            self.installation.update(package, version)
            installation_changed = True
        if installation_changed:
            logging.info('Package upgraded successfully!')
            self.installation.setup()

    def update_all(self, command_args):
        parser = argparse.ArgumentParser(prog='{0} upgrade-all'.format(CLI_NAME),
                                         description='Update all packages to latest release/version.')
        parser.add_argument('--yes', help='continue without confirmation', action='store_true')
        args = parser.parse_args(command_args)
        if not args.yes:
            logging.info('All packages will be upgraded to latest versions.')
            logging.info('Continue? [Y/n]?')
            if not raw_input().strip().lower().startswith('y'):
                sys.exit(0)
        for package in self.installation.updated_packages():
            for version in package.updated_versions:
                self.installation.update(package, version)
        logging.info('All packages upgraded successfully!')
        self.installation.setup()

    def rebuild(self, command_args):
        parser = argparse.ArgumentParser(prog='{0} rebuild'.format(CLI_NAME), description='Rebuild installation.')
        parser.parse_args(command_args)
        logging.warning('Rebuilding whole installation!')
        logging.warning('This might take a LONG time depending on number of missing/damaged packages!')
        self.installation.rebuild()
        self.installation.setup()
        logging.info('Rebuild completed successfuly')

    def grids(self, command_args):
        parser = argparse.ArgumentParser(prog='{0} grids'.format(CLI_NAME), description='List available grids.')
        parser.parse_args(command_args)
        for grid in self.installation.grids:
            logging.debug((grid.lower(), CLI_NAME))
            if grid.lower() == CLI_NAME:
                logging.info(green(grid))
            else:
                logging.info(grid)

    def collections(self, command_args):
        parser = argparse.ArgumentParser(prog='{0} collections'.format(CLI_NAME), description='List available collections.')
        parser.parse_args(command_args)
        for grid in self.installation.grids:
            logging.debug((grid.lower(), CLI_NAME))
            if grid.lower() == CLI_NAME:
                logging.info(green(grid))
            else:
                logging.info(grid)
            for collection in self.installation.collections[grid]:
                logging.info(' - {0}'.format(collection))

    def reactivate(self, command_args):
        def is_sqlite(filename):
            with open(filename, 'r') as packages_file:
                if packages_file.read(15) == 'SQLite format 3':
                    return True
            return False

        parser = argparse.ArgumentParser(prog='{0} reactivate'.format(CLI_NAME),
                                         description='Recreate installation based on config and package list.')
        parser.add_argument('config', help='path to config file')
        parser.add_argument('packages', nargs='?', help='path to packages database or list', default=None)
        args = parser.parse_args(command_args)
        if not os.path.exists(args.config):
            raise Exception('Invalid path to config file!')
        if args.packages is not None and not os.path.exists(args.packages):
            raise Exception('Invalid path to packages list!')
        config = ConfigParser()
        config.read(args.config)
        if not config.has_section('installer'):
            raise NotActivatedException('Invalid config file!')
        logging.info('Activating installation of site: %s for user: %s',
                     config.get('installer', 'site'),
                     config.get('installer', 'user'))
        if not self.installation.activate(config.get('installer', 'site'),
                                          config.get('installer', 'user'),
                                          config.get('installer', 'key')):
            raise Exception('Activation failed!')
        logging.info('Activation successful!')
        rsync_host = ''
        rsync_port = ''
        if config.has_option('installer', 'rsync_custom_host'):
            rsync_host = config.get('installer', 'rsync_custom_host')
        if config.has_option('installer', 'rsync_custom_port'):
            rsync_port = config.get('installer', 'rsync_custom_port')
        self.installation.save_activation_info(rsync_host, rsync_port)
        if config.has_section('additional_params') and config.get('additional_params', 'multiarch'):
            platforms = self.installation.save_multiarch_config()
            logging.info('Multibranch installation mode is on')
            logging.info('List of available architectures: ')
            for platform_name in platforms:
                logging.info(' - {0}'.format(platform_name))
        logging.info('Updating to latest version ...')
        self.bootstrap_installer()
        packages = list()
        if config.has_section('packages'):
            for key, value in config.items('packages'):
                package_version_array = key.split('@')
                packages.append((package_version_array[0], package_version_array[1]))
        elif is_sqlite(args.packages):
            conn = sqlite3.connect(args.packages)
            for slug, ver, default in conn.execute('SELECT slug, version, default_version FROM packages'):
                packages.append((slug, None if default else ver))
        else:
            # Assume this is a text file
            with open(args.packages, 'r') as f:
                for line in f.readlines():
                    if line.startswith('#'):
                        continue
                    line = line.strip()
                    if '==' in line:
                        slug, ver = line.split('==')
                    else:
                        slug, ver = line, None
                    packages.append((slug, ver))
        for slug, ver in packages:
            package = self.installation.package(slug)
            if ver:
                version = package.version(ver)
            else:
                version = package.default_version
            logging.info('Installing: {0}@{1}'.format(package.slug, version.version))
            self.installation.install(package, version, version.os_dir)
        if packages:
            self.installation.setup()
        logging.info('Reactivation completed.')


def main():
    Installer()


if __name__ == '__main__':
    main()
